from dataclasses import dataclass

from dateutil.parser import parse


@dataclass
class ThreadMember:
    # Attrs
    def __init__(self, json) -> None:
        for key in json:
            if key == "join_timestamp":
                setattr(self, key, parse(json[key]) if json[key] else None)
            else:
                setattr(self, key, json[key])



class GetChannelError(Exception):
    """Raises when fetch channel is failed."""

    pass


class SendMessageToChannelFailed(Exception):
    """Raises when send message to channel is failed."""

    pass


class EditChannelFailed(Exception):
    """Raises when editing the channel is failed."""

    pass


class DeleteChannelFailed(Exception):
    """Raises when deleting the channel is failed."""

    pass


class FetchChannelHistoryFailed(Exception):
    """Raises when fetching the channel history is failed."""

    pass


class FetchChannelMessageFailed(Exception):
    """Raises when fetching the channel message is failed."""

    pass


class FetchChannelInvitesFailed(Exception):
    """Raises when fetching the channel invites is failed."""

    pass


class CreateInviteFailed(Exception):
    """Raises when creating new invite is failed."""

    pass


class FetchPinnedMessagesFailed(Exception):
    """Raises when fetching the pinned messages is failed."""

    pass


class PinMessageFailed(Exception):
    """Raises when pinning a message is failed."""

    pass


class UnpinMessageFailed(Exception):
    """Raises when unpinning a message is failed."""

    pass


class EditChannelPermissionsFailed(Exception):
    """Raises when editing the channel permissions is failed."""

    pass


class DeleteChannelPermissionsFailed(Exception):
    """Raises when deleting the channel permissions is failed."""

    pass


class TriggerTypingFailed(Exception):
    """Raises when trigger the typing is failed."""

    pass



from dataclasses import dataclass
from datetime import datetime


@dataclass
class Embed:
    # Attrs
    def __init__(self, data=None) -> None:
        if data is None:
            data = {}
        self.color = None
        self.title = None
        self.description = None
        self.url = None
        self.author = None
        self.image = None
        self.thumbnail = None
        self.timestamp = None
        self.fields = None
        self.footer = None

        for key in data:
            setattr(self, key, data[key])

    def __call__(self):
        self.embed = {}

        if self.title:
            self.embed['title'] = self.title

        if self.url:
            self.embed['url'] = self.url

        if self.color:
            self.embed['color'] = self.color

        if self.description:
            self.embed['description'] = self.description

        if self.author:
            self.embed['author'] = self.author

        if self.thumbnail:
            self.embed['thumbnail'] = self.thumbnail

        if self.image:
            self.embed['image'] = self.image

        if self.timestamp:
            self.embed['timestamp'] = self.timestamp

        if self.footer:
            self.embed['footer'] = self.footer

        if self.footer:
            self.embed['footer'] = self.footer

        if self.fields:
            self.embed['fields'] = self.fields

        return self.embed

    def set_author(self, **kwargs):
        """Set author to the embed."""

        self.author = {}

        if 'text' in kwargs:
            self.author['text'] = kwargs['name']

        if 'icon_url' in kwargs:
            self.author['icon_url'] = kwargs['icon_url']

        if 'url' in kwargs:
            self.author['url'] = kwargs['url']

    def set_footer(self, **kwargs):
        """Set footer to the embed."""

        self.footer = {}

        if 'text' in kwargs:
            self.footer['text'] = kwargs['text']

        if 'icon_url' in kwargs:
            self.footer['icon_url'] = kwargs['icon_url']

    def set_image(self, url: str):
        """Set image to the embed."""

        self.image = {"url": url}

    def set_thumbnail(self, url: str):
        """Set thumbnail to the embed."""

        self.thumbnail = {"url": url}

    def set_timestamp(self):
        """Add current timestamp to the embed."""

        self.timestamp = datetime.utcnow().isoformat()

    def add_field(self, name: str, value: str, inline: bool = True):
        """Add new field to the embed."""

        if not self.fields:
            self.fields = []

        self.fields.append({
            "name": name,
            "value": value,
            "inline": inline
        })



class TokenNotFoundError(Exception):
    """Raises when token not found in session object."""

    pass


class IntentNotFoundError(Exception):
    """Raises when token not found in session object."""

    pass


class InvalidTokenError(Exception):
    """Raises when token is invalid in session object."""

    pass


class FetchInviteFailedError(Exception):
    """Raises when fetcing the invite is failed."""

    pass


class RemoveInviteFailedError(Exception):
    """Raises when removing the invite is failed."""

    pass


class FetchUserFailedError(Exception):
    """Raises when fetching the user is failed."""

    pass


class FetchGuildFailedError(Exception):
    """Raises when fetching the guild is failed."""

    pass


class EditClientUserFailed(Exception):
    """Raises when editing the current user is failed."""

    pass


class LeaveGuildFailed(Exception):
    """Raises when leaving the guild is failed."""

    pass


class FetchGuildPreviewFailed(Exception):
    """Raises when fetching the guild preview is failed."""

    pass



from dataclasses import dataclass
from typing import Union


@dataclass
class RoleTags:
    # Attrs
    def __init__(self, json) -> None:
        self.bot_id: Union[str,
                           None] = json['bot_id'] if 'bot_id' in json else None
        self.integration_id: Union[str,
                                   None] = json['integration_id'] if 'integration_id' in json else None
        self.premium_subscriber: None = json['premium_subscriber'] if 'premium_subscriber' in json else None



"""
GUILDS (1 << 0)
  - GUILD_CREATE
  - GUILD_UPDATE
  - GUILD_DELETE
  - GUILD_ROLE_CREATE
  - GUILD_ROLE_UPDATE
  - GUILD_ROLE_DELETE
  - CHANNEL_CREATE
  - CHANNEL_UPDATE
  - CHANNEL_DELETE
  - CHANNEL_PINS_UPDATE
  - THREAD_CREATE
  - THREAD_UPDATE
  - THREAD_DELETE
  - THREAD_LIST_SYNC
  - THREAD_MEMBER_UPDATE
  - THREAD_MEMBERS_UPDATE
  - STAGE_INSTANCE_CREATE
  - STAGE_INSTANCE_UPDATE
  - STAGE_INSTANCE_DELETE

GUILD_MEMBERS (1 << 1)
  - GUILD_MEMBER_ADD
  - GUILD_MEMBER_UPDATE
  - GUILD_MEMBER_REMOVE
  - THREAD_MEMBERS_UPDATE

GUILD_BANS (1 << 2)
  - GUILD_BAN_ADD
  - GUILD_BAN_REMOVE

GUILD_EMOJIS (1 << 3)
  - GUILD_EMOJIS_UPDATE

GUILD_INTEGRATIONS (1 << 4)
  - GUILD_INTEGRATIONS_UPDATE
  - INTEGRATION_CREATE
  - INTEGRATION_UPDATE
  - INTEGRATION_DELETE

GUILD_WEBHOOKS (1 << 5)
  - WEBHOOKS_UPDATE

GUILD_INVITES (1 << 6)
  - INVITE_CREATE
  - INVITE_DELETE

GUILD_VOICE_STATES (1 << 7)
  - VOICE_STATE_UPDATE

GUILD_PRESENCES (1 << 8)
  - PRESENCE_UPDATE

GUILD_MESSAGES (1 << 9)
  - MESSAGE_CREATE
  - MESSAGE_UPDATE
  - MESSAGE_DELETE
  - MESSAGE_DELETE_BULK

GUILD_MESSAGE_REACTIONS (1 << 10)
  - MESSAGE_REACTION_ADD
  - MESSAGE_REACTION_REMOVE
  - MESSAGE_REACTION_REMOVE_ALL
  - MESSAGE_REACTION_REMOVE_EMOJI

GUILD_MESSAGE_TYPING (1 << 11)
  - TYPING_START

DIRECT_MESSAGES (1 << 12)
  - MESSAGE_CREATE
  - MESSAGE_UPDATE
  - MESSAGE_DELETE
  - CHANNEL_PINS_UPDATE

DIRECT_MESSAGE_REACTIONS (1 << 13)
  - MESSAGE_REACTION_ADD
  - MESSAGE_REACTION_REMOVE
  - MESSAGE_REACTION_REMOVE_ALL
  - MESSAGE_REACTION_REMOVE_EMOJI

DIRECT_MESSAGE_TYPING (1 << 14)
  - TYPING_START
"""

GUILDS = 0x0000000001
GUILD_MEMBERS = 0x0000000002
GUILD_BANS = 0x0000000004
GUILD_EMOJIS = 0x0000000008
GUILD_INTEGRATIONS = 0x0000000010
GUILD_WEBHOOKS = 0x0000000020
GUILD_INVITES = 0x0000000040
GUILD_VOICE_STATES = 0x0000000080
GUILD_PRESENCES = 0x0000000100
GUILD_MESSAGES = 0x0000000200
GUILD_MESSAGE_REACTIONS = 0x0000000400
GUILD_MESSAGE_TYPING = 0x0000000800
DIRECT_MESSAGES = 0x0000001000
DIRECT_MESSAGE_REACTIONS = 0x0000002000
DIRECT_MESSAGE_TYPING = 0x0000004000
ALL = 0x7FFF



from dataclasses import dataclass


@dataclass
class Reaction:
    # Attrs
    def __init__(self, json, token) -> None:
        self.__token: str = token

        from .Emoji import Emoji

        self.count: int = json['count']
        self.reacted: bool = json['me']
        self.emoji: Emoji = Emoji(json['emoji'], self.__token)



from base64 import b64encode


def open_and_convert(path: str):
    with open(path, "rb") as f:
        img_bytes = f.read()
        encrypt = b64encode(img_bytes).decode()

        return f"data:image/{f.name.split('.')[-1].replace('jpg', 'jpeg')};base64,{encrypt}"


def convert(data: bytes, img_type: str):
    encrypt = b64encode(data).decode()

    return f"data:image/{img_type};base64,{encrypt}"



from dataclasses import dataclass
from urllib.parse import quote

from dateutil.parser import parse

from ..errors.ChannelErrors import *
from ..errors.MessageErrors import *
from ..utils.Dict2Query import convert as d2q_converter
from ..utils.Request import Request
from ..utils.WrongType import raise_error


@dataclass
class Message:
    # Attrs
    def __init__(self, json, token) -> None:
        self.id = None
        self.channel_id = None
        from .Reaction import Reaction
        from .MessageActivity import MessageActivity
        from .Application import Application
        from .MessageReference import MessageReference
        from .Sticker import Sticker
        from .Channel import Channel
        from .Member import Member
        from .User import User
        from .Embed import Embed
        from .Attachment import Attachment

        self.__token: str = token

        for key in json:
            if key in ("mentions", "author"):
                setattr(
                    self,
                    key,
                    User(
                        json[key],
                        self.__token) if key == "author" else [
                        User(
                            i,
                            self.__token) for i in json[key]])
            elif key == "member":
                setattr(self, key, Member(json[key], self.__token))
            elif key in ("timestamp", "edited_timestamp"):
                setattr(self, key, parse(json[key]) if json[key] else None)
            elif key == "mentions":
                setattr(self, key, [User(i, self.__token) for i in json[key]])
            elif key == "attachments":
                setattr(self, key, [Attachment(i) for i in json[key]])
            elif key == "embeds":
                setattr(self, key, [Embed(i) for i in json[key]])
            elif key == "reactions":
                setattr(self, key, [Reaction(i, self.__token)
                                    for i in json[key]])
            elif key == "activity":
                setattr(self, key, MessageActivity(json[key]))
            elif key == "application":
                setattr(self, key, Application(json[key], self.__token))
            elif key == "message_reference":
                setattr(self, key, MessageReference(json[key]))
            elif key == "stickers":
                setattr(self, key, [Sticker(i) for i in json[key]])
            elif key == "referenced_message":
                setattr(
                    self,
                    key,
                    Message(
                        json[key],
                        self.__token) if json[key] else None)
            elif key == "thread":
                setattr(self, key, Channel(json[key], self.__token))
            else:
                setattr(self, key, json[key])

    async def reply(self, options=None):
        """Reply to the message with API params."""
        if options is None:
            options = {}
        raise_error(options, "options", dict)

        atom, result = await Request().send_async_request(f"/channels/{self.channel_id}/messages", "POST", self.__token,
                                                          {
                                                              **options,
                                                              "message_reference": {
                                                                  "message_id": self.id
                                                              }
                                                          })

        if atom == 0:
            return Message(result, self.__token)
        else:
            raise SendMessageToChannelFailed(result)

    async def edit(self, options=None):
        """Edit your message with API params."""
        if options is None:
            options = {}
        raise_error(options, "options", dict)

        atom, result = await Request().send_async_request(f"/channels/{self.channel_id}/messages/{self.id}", "PATCH",
                                                          self.__token, options)

        if atom == 0:
            return Message(result, self.__token)
        else:
            raise EditMessageFailed(result)

    async def delete(self):
        """Delete the message."""

        atom, result = await Request().send_async_request(f"/channels/{self.channel_id}/messages/{self.id}", "DELETE",
                                                          self.__token)

        if atom == 0:
            return self
        else:
            raise DeleteMessageFailed(result)

    async def add_reaction(self, emoji: str):
        """Add reaction to message."""
        raise_error(emoji, "emoji", str)

        atom, result = await Request().send_async_request(
            f"/channels/{self.channel_id}/messages/{self.id}/reactions/{quote(emoji)}/@me", "PUT", self.__token)

        if atom == 0:
            return self
        else:
            raise AddReactionToMessageFailed(result)

    async def remove_reaction(self, emoji: str, user: str = None):
        """Remove an user reaction from message."""
        raise_error(emoji, "emoji", str)

        if user is not None:
            raise_error(user, "user", str)

        atom, result = await Request().send_async_request(
            f"/channels/{self.channel_id}/messages/{self.id}/reactions/{quote(emoji)}/{'@me' if user is None else user}",
            "DELETE", self.__token)

        if atom == 0:
            return self
        else:
            raise RemoveReactionToMessageFailed(result)

    async def fetch_reactions(self, emoji: str, options=None):
        """Fetch message reactions with API params."""

        if options is None:
            options = {"limit": 25}
        from .User import User

        raise_error(emoji, "emoji", str)
        raise_error(options, "options", dict)

        query_param = f"/channels/{self.channel_id}/messages/{self.id}/reactions/{emoji}{d2q_converter(options)}"

        atom, result = await Request().send_async_request(query_param, "GET", self.__token)

        if atom == 0:
            return [User(i, self.__token) for i in result]
        else:
            raise FetchReactionsFromMessageFailed(result)

    async def remove_reactions(self, emoji: str = None):
        """Fetch message reactions with API params."""
        if emoji is not None:
            raise_error(emoji, "emoji", str)

        atom, result = await Request().send_async_request(
            f"/channels/{self.channel_id}/messages/{self.id}/reactions{'/' + quote(emoji) if emoji is not None else ''}",
            "DELETE", self.__token)

        if atom == 0:
            return True
        else:
            raise RemoveReactionsFromMessageFailed(result)

    async def crosspost(self):
        """Cross post the message (https://discord.com/developers/docs/resources/channel#crosspost-message)"""

        atom, result = await Request().send_async_request(f"/channels/{self.channel_id}/messages/{self.id}/crosspost",
                                                          "POST", self.__token)

        if atom == 0:
            return Message(result, self.__token)
        else:
            raise CrossPostMessageFailed(result)



from dataclasses import dataclass

from dateutil.parser import parse


@dataclass
class Invite:
    # Attrs
    def __init__(self, json, token) -> None:
        self.__token = token

        from .Channel import Channel
        from .User import User
        from .Guild import Guild
        from .Application import Application

        for key in json:
            if key == "guild":
                setattr(self, key, Guild(json[key], self.__token))
            elif key == "channel":
                setattr(self, key, Channel(json[key], self.__token))
            elif key in ("inviter", "target_user",):
                setattr(self, key, User(json[key], self.__token))
            elif key == "target_application":
                setattr(self, key, Application(json[key], self.__token))
            elif key == "expires_at":
                setattr(self, key, parse(json[key]) if json[key] else None)
            else:
                setattr(self, key, json[key])



from dataclasses import dataclass


@dataclass
class Role:
    # Attrs
    def __init__(self, json) -> None:
        from .RoleTags import RoleTags

        for key in json:
            if key == "tags":
                setattr(self, key, RoleTags(json[key]))
            else:
                setattr(self, key, json[key])



from dataclasses import dataclass

from ..errors.ChannelErrors import *
from ..errors.MessageErrors import *
from ..errors.SessionErrors import *
from ..utils.Dict2Query import convert as d2q_converter
from ..utils.Request import Request
from ..utils.WrongType import raise_error


@dataclass
class ClientUser:
    # Attrs
    def __init__(self, json: dict, token: str, max_messages: int) -> None:
        self.__token: str = token
        self.max_messages = max_messages

        for key in json:
            setattr(self, key, json[key])

        self.channels: list = []

        self.cache: dict = {
            "message": [],
            "guild": []
        }

    async def add_message_cache(self, message):
        """Add message to cache"""

        self.cache['message'].append(message)
        
        if len(self.cache['message']) > self.max_messages:
            del self.cache['message'][0]

    async def remove_message_cache(self, packet):
        """Remove deleted message from cache"""

        for index, cache in enumerate(self.cache['message']):
            if cache.id == packet['id'] and cache.channel_id == packet['channel_id']:
                del self.cache['message'][index]
                break

    async def bulk_delete_message_cache(self, packet):
        """Remove Bulk-deleted message from cache"""

        shift = 0
        for index, cache in enumerate(self.cache['message'][:]):
            if cache.id in packet['ids'] and cache.channel_id == packet['channel_id']:
                del self.cache['message'][index - shift]
                shift += 1

    async def update_message_cache(self, message):
        """Update message from cache"""

        for index, cache in enumerate(self.cache['message']):
            if cache.id == message.id and cache.channel_id == message.channel_id:
                self.cache['message'][index] = message
                break

    async def add_guild_cache(self, guild):
        """Add a guild to cache."""

        self.cache["guild"].append(guild)

    async def update_guild_cache(self, guild):
        """Update a guild from cache."""

        for index, cache in enumerate(self.cache['guild']):
            if cache.id == guild.id:
                self.cache['guild'][index] = guild
                break

    async def remove_guild_cache(self, packet):
        """Remove a guild from cache."""

        for index, cache in enumerate(self.cache['guild'][:]):
            if cache.id == packet['id']:
                del self.cache['guild'][index]
                break

    async def add_guild_member(self, guild_id, member):
        """Add a guild member to guild cache."""

        for index, cache in enumerate(self.cache['guild']):
            if cache.id == guild_id:
                self.cache['guild'][index].members.append(member)
                break

    async def remove_guild_member(self, guild_id, user):
        """Remove a guild member from guild cache."""

        for index, cache in enumerate(self.cache['guild']):
            if cache.id == guild_id:
                for user_index, cache_user in enumerate(cache.members):
                    if cache_user.user.id == user.id:
                        self.cache['guild'][index].members.pop(user_index)
                        break
                break

    async def update_guild_member(self, guild_id, member):
        """Update a guild member from guild cache."""

        for index, cache in enumerate(self.cache['guild']):
            if cache.id == guild_id:
                for user_index, cache_user in enumerate(cache.members):
                    if cache_user.user.id == member.user.id:
                        self.cache['guild'][index].members[user_index] = member
                        break
                break

    async def add_guild_channel(self, channel):
        """Add a guild channel to guild cache."""

        for index, cache in enumerate(self.cache['guild']):
            if hasattr(channel, "guild_id") and cache.id == channel.guild_id:
                self.cache['guild'][index].channels.append(channel)
                break

    async def remove_guild_channel(self, channel):
        """Remove a guild channel from guild cache."""

        for index, cache in enumerate(self.cache['guild']):
            if hasattr(channel, "guild_id") and cache.id == channel.guild_id:
                for channel_index, cache_channel in enumerate(cache.channels):
                    if cache_channel.id == channel.id:
                        self.cache['guild'][index].channels.pop(channel_index)
                        break
                break

    async def update_guild_channel(self, channel):
        """Update a guild channel from guild cache."""

        for index, cache in enumerate(self.cache['guild']):
            if hasattr(channel, "guild_id") and cache.id == channel.guild_id:
                for channel_index, cache_channel in enumerate(cache.channels):
                    if cache_channel.id == channel.id:
                        self.cache['guild'][index].channels[channel_index] = channel
                        break
                break

    async def get_channel(self, channel_id: str, fetch: bool = False):
        """Get channel with id."""
        raise_error(channel_id, "id", str)
        raise_error(fetch, "fetch", bool)

        from .Channel import Channel
        for channel in self.channels:
            if channel_id == channel.id and not fetch:
                return channel

        atom, result = await Request().send_async_request(f"/channels/{channel_id}", "GET", self.__token)

        if atom == 0:
            channel_object = Channel(result, self.__token)
            self.channels.append(channel_object)

            return channel_object
        else:
            raise GetChannelError(f"ATOM #{atom}: #{result}")

    async def bulk_delete(self, channel_id: str, limit: int = 10):
        """Bulk-delete channel messages with channel id."""
        raise_error(channel_id, "id", str)
        raise_error(limit, "limit", int)

        from .Message import Message

        # filtered = [i.id for i in self.cache["message"] if i.channel_id == channel_id][::-1][:limit]

        # if len(filtered) < 2:
        query_format = f"/channels/{channel_id}/messages?limit={limit}"
        atom, result = await Request().send_async_request(query_format, "GET", self.__token)

        if atom == 0:
            filtered = [
                           Message(i, self.__token).id for i in result][::-1][:limit]
        else:
            raise FetchChannelHistoryFailed(result)

        atom, result = await Request().send_async_request(f"/channels/{channel_id}/messages/bulk-delete", "POST",
                                                          self.__token, {"messages": filtered})

        if atom == 0:
            return filtered
        else:
            raise BulkDeleteMessageFailed(result)

    async def say(self, channel_id: str, params=None):
        """Send message to the channel."""
        if params is None:
            params = {}
        raise_error(channel_id, "channel_id", str)

        if params is not None:
            raise_error(params, "params", dict)

        from .Message import Message

        atom, result = await Request().send_async_request(f"/channels/{channel_id}/messages", "POST", self.__token,
                                                          params)

        if atom == 0:
            return Message(result, self.__token)
        else:
            raise SendMessageToChannelFailed(result)

    async def fetch_invite(self, code: str, options=None):
        """Get invite information with invite code (supports API params.)."""
        if options is None:
            options = {}
        raise_error(code, "code", str)

        if options is not None:
            raise_error(options, "options", dict)

        from .Invite import Invite

        atom, result = await Request().send_async_request(f"/invites/{code}{d2q_converter(options)}", "GET",
                                                          self.__token)

        if atom == 0:
            return Invite(result, self.__token)
        else:
            raise FetchInviteFailedError(result)

    async def remove_invite(self, code: str):
        """Remove invite with invite code."""
        raise_error(code, "code", str)

        from .Invite import Invite

        atom, result = await Request().send_async_request(f"/invites/{code}", "DELETE", self.__token)

        if atom == 0:
            return Invite(result, self.__token)
        else:
            raise RemoveInviteFailedError(result)

    async def fetch_client_user(self):
        """Fetch information about client user."""
        from .User import User

        atom, result = await Request().send_async_request("/users/@me", "GET", self.__token)

        if atom == 0:
            return User(result, self.__token)
        else:
            raise FetchUserFailedError(result)

    async def fetch_user(self, user_id: str):
        """Fetch information about an user."""
        raise_error(user_id, "user_id", str)

        from .User import User

        atom, result = await Request().send_async_request(f"/users/{user_id}", "GET", self.__token)

        if atom == 0:
            return User(result, self.__token)
        else:
            raise FetchUserFailedError(result)

    def get_guild(self, guild_id: str):
        """Get guild from cache."""
        raise_error(guild_id, "id", str)

        for guild in self.cache["guild"]:
            if guild.id == guild_id:
                return guild

    async def fetch_guild(self, guild_id: str, options=None):
        """Fetch guild with API params."""
        if options is None:
            options = {}
        raise_error(guild_id, "id", str)
        raise_error(options, "options", dict)

        from .Guild import Guild

        atom, result = await Request().send_async_request(f"/guilds/{guild_id}{d2q_converter(options)}", "GET",
                                                          self.__token)

        if atom == 0:
            return Guild(result, self.__token)
        else:
            raise FetchGuildFailedError(result)

    async def edit_user(self, params=None):
        """Modify current user with API params."""
        if params is None:
            params = {}
        raise_error(params, "params", dict)

        from .User import User

        atom, result = await Request().send_async_request("/users/@me", "PATCH", self.__token, params)

        if atom == 0:
            return User(result, self.__token)
        else:
            raise EditClientUserFailed(result)

    async def leave_guild(self, guild_id: str):
        """Leave a guild with id."""
        raise_error(guild_id, "guild_id", str)

        atom, result = await Request().send_async_request(f"/users/@me/guilds/{guild_id}", "DELETE", self.__token, {})

        if atom == 0:
            return True
        else:
            raise LeaveGuildFailed(result)

    async def fetch_guild_preview(self, guild_id: str):
        """Fetch guild preview with id."""
        raise_error(guild_id, "guild_id", str)

        from .GuildPreview import GuildPreview

        atom, result = await Request().send_async_request(f"/guilds/{guild_id}/preview", "GET", self.__token)

        if atom == 0:
            return GuildPreview(result, self.__token)
        else:
            raise FetchGuildPreviewFailed(result)



from dataclasses import dataclass

from dateutil.parser import parse


@dataclass
class VoiceState:
    # Attrs
    def __init__(self, json, token) -> None:
        self.__token: str = token

        from .Member import Member

        for key in json:
            if key == "member":
                setattr(self, key, Member(json[key], self.__token))
            elif key == "request_to_speak_timestamp":
                setattr(self, key, parse(json[key]) if json[key] else None)
            else:
                setattr(self, key, json[key])



from dataclasses import dataclass
from typing import Union

from aiohttp import ClientSession, client_exceptions


@dataclass
class Request(object):
    api: str = "https://discord.com/api/v9"

    async def send_async_request(self, endpoint: str, method: str, token: str, body: Union[dict, list] = None) -> tuple:
        """Send an async request to discord API."""

        result: tuple = ()
        # Atom 0: When Request Success (atom, json)
        # Atom 1: When Request Got Error (atom, error_message)

        url: str = f"{self.api}{endpoint}"

        async with ClientSession(trust_env=True) as session:
            async with session.request(method, url,
                                       headers={"Authorization": f"Bot {token}", 'Content-Type': 'application/json'},
                                       json=body) as response:
                try:
                    json_data = await response.json()
                except client_exceptions.ContentTypeError:
                    body_text = await response.text()

                    if str(response.status).startswith("2"):
                        return (0, "")
                    else:
                        return (1, body_text)

                if not str(response.status).startswith("2"):
                    result = (
                        1,
                        f"Error ({response.status}): {json_data['message']}\nRetry After? {json_data['retry_after'] if 'retry_after' in json_data else 'Not Found'}",
                    )
                else:
                    result = (0, json_data,)

                return result



from dataclasses import dataclass
from datetime import datetime
from typing import Union


@dataclass
class Activity:
    # Attrs
    def __init__(self, json) -> None:
        self.name: str = json['name']
        self.type: int = json['type']
        self.url: Union[str, None] = json['url'] if 'url' in json else None
        self.created_at: datetime = datetime.fromtimestamp(json['created_at'])



class EditMessageFailed(Exception):
    """Raises when editing the message is failed."""

    pass


class DeleteMessageFailed(Exception):
    """Raises when deleting the message is failed."""

    pass


class BulkDeleteMessageFailed(Exception):
    """Raises when bulk-delete the messages is failed."""

    pass


class AddReactionToMessageFailed(Exception):
    """Raises when adding a reaction to message is failed."""

    pass


class RemoveReactionToMessageFailed(Exception):
    """Raises when removing a reaction to message is failed."""

    pass


class FetchReactionsFromMessageFailed(Exception):
    """Raises when fetching reactions from message is failed."""

    pass


class RemoveReactionsFromMessageFailed(Exception):
    """Raises when removing reactions from message is failed."""

    pass


class CrossPostMessageFailed(Exception):
    """Raises when cross posting a message is failed."""

    pass



from dataclasses import dataclass

from dateutil.parser import parse

from ..errors.ChannelErrors import *
from ..utils.Dict2Query import convert as d2q_converter
from ..utils.Request import Request
from ..utils.WrongType import raise_error


@dataclass
class Channel:
    # Attrs
    def __init__(self, json: dict, token: str) -> None:
        self.id = None

        from .Overwrite import Overwrite
        from .User import User
        from .ThreadMetadata import ThreadMetadata
        from .ThreadMember import ThreadMember

        self.__token: str = token

        for key in json:
            if key == "permission_overwrites":
                setattr(self, key, [Overwrite(i) for i in json[key]])
            elif key == "recipients":
                setattr(self, key, [User(i, self.__token) for i in json[key]])
            elif key == "last_pin_timestamp":
                setattr(self, key, parse(json[key]) if json[key] else None)
            elif key == "thread_metadata":
                setattr(self, key, ThreadMetadata(json[key]))
            elif key == "member":
                setattr(self, key, ThreadMember(json[key]))
            else:
                setattr(self, key, json[key])

    async def send(self, options=None):
        """Send message to the channel."""
        if options is None:
            options = {}

        raise_error(options, "options", dict)

        from .Message import Message

        atom, result = await Request().send_async_request(f"/channels/{self.id}/messages", "POST", self.__token,
                                                          options)

        if atom == 0:
            return Message(result, self.__token)
        else:
            raise SendMessageToChannelFailed(result)

    async def edit(self, options=None):
        """Edit channel with API params."""
        if options is None:
            options = {}
        raise_error(options, "options", dict)

        atom, result = await Request().send_async_request(f"/channels/{self.id}", "PATCH", self.__token, options)

        if atom == 0:
            return Channel(result, self.__token)
        else:
            raise EditChannelFailed(result)

    async def delete(self):
        """Delete current channel."""

        atom, result = await Request().send_async_request(f"/channels/{self.id}", "DELETE", self.__token, {})

        if atom == 0:
            return Channel(result, self.__token)
        else:
            raise DeleteChannelFailed(result)

    async def fetch_history(self, options=None):
        """Fetch channel history with API params."""
        if options is None:
            options = {}
        raise_error(options, "options", dict)

        from .Message import Message

        atom, result = await Request().send_async_request(f"/channels/{self.id}/messages{d2q_converter(options)}",
                                                          "GET", self.__token)

        if atom == 0:
            return [Message(i, self.__token) for i in result]
        else:
            raise FetchChannelHistoryFailed(result)

    async def fetch_message(self, message_id: str):
        """Fetch channel message with id."""
        raise_error(message_id, "id", str)

        from .Message import Message

        atom, result = await Request().send_async_request(f"/channels/{self.id}/messages/{message_id}", "GET",
                                                          self.__token)

        if atom == 0:
            return Message(result, self.__token)
        else:
            raise FetchChannelMessageFailed(result)

    async def fetch_invites(self):
        """Fetch channel invites."""
        from .Invite import Invite

        atom, result = await Request().send_async_request(f"/channels/{self.id}/invites", "GET", self.__token)

        if atom == 0:
            return [Invite(i, self.__token) for i in result]
        else:
            raise FetchChannelInvitesFailed(result)

    async def create_invite(self, params=None):
        """Create new invite with API params."""
        if params is None:
            params = {}
        raise_error(params, "params", dict)

        from .Invite import Invite

        atom, result = await Request().send_async_request(f"/channels/{self.id}/invites", "POST", self.__token, params)

        if atom == 0:
            return Invite(result, self.__token)
        else:
            raise CreateInviteFailed(result)

    async def pinned_messages(self):
        """Fetch pinned messages."""
        from .Message import Message

        atom, result = await Request().send_async_request(f"/channels/{self.id}/pins", "GET", self.__token)

        if atom == 0:
            return [Message(i, self.__token) for i in result]
        else:
            raise FetchPinnedMessagesFailed(result)

    async def pin_message(self, message_id: str):
        """Pin a message with id."""
        raise_error(message_id, "message_id", str)

        atom, result = await Request().send_async_request(f"/channels/{self.id}/pins/{message_id}", "PUT", self.__token)

        if atom == 0:
            return True
        else:
            raise PinMessageFailed(result)

    async def unpin_message(self, message_id: str):
        """Unpin a message with id."""
        raise_error(message_id, "message_id", str)

        atom, result = await Request().send_async_request(f"/channels/{self.id}/pins/{message_id}", "DELETE",
                                                          self.__token, {})

        if atom == 0:
            return True
        else:
            raise UnpinMessageFailed(result)

    async def edit_permissions(self, user_or_role_id: str, params: dict):
        """Edit channel permission. (bitwise must be string.)"""
        raise_error(user_or_role_id, "user_or_role_id", str)
        raise_error(params, "params", dict)

        atom, result = await Request().send_async_request(f"/channels/{self.id}/permissions/{user_or_role_id}", "PUT",
                                                          self.__token, params)

        if atom == 0:
            return self
        else:
            raise EditChannelPermissionsFailed(result)

    async def delete_permissions(self, user_or_role_id: str):
        """Delete permissions from role/user"""
        raise_error(user_or_role_id, "user_or_role_id", str)

        atom, result = await Request().send_async_request(f"/channels/{self.id}/permissions/{user_or_role_id}",
                                                          "DELETE", self.__token, {})

        if atom == 0:
            return self
        else:
            raise DeleteChannelPermissionsFailed(result)

    async def trigger_typing(self):
        """Start typing in this channel."""

        atom, result = await Request().send_async_request(f"/channels/{self.id}/typing", "POST", self.__token)

        if atom == 0:
            return self
        else:
            raise TriggerTypingFailed(result)



from dataclasses import dataclass


@dataclass
class User:
    # Attrs
    def __init__(self, json, token) -> None:
        self.id = None
        self.discriminator = None
        self.username = None
        self.__token: str = token

        for key in json:
            setattr(self, key, json[key])

    @property
    def tag(self) -> str:
        return f"{self.username}#{self.discriminator}"

    @property
    def mention(self) -> str:
        return f"<@{self.id}>"



from dataclasses import dataclass

from dateutil.parser import parse

from ..errors.GuildErrors import *
from ..utils.Dict2Query import convert as d2q_converter
from ..utils.Request import Request
from ..utils.WrongType import raise_error


@dataclass
class Guild:
    # Attrs
    def __init__(self, json, token) -> None:
        self.id = None
        self.__token = token

        from .Role import Role
        from .Emoji import Emoji
        from .VoiceState import VoiceState
        from .Member import Member
        from .Channel import Channel

        for key in json:
            if key == "roles":
                setattr(self, key, [Role(i) for i in json[key]])
            elif key == "emojis":
                setattr(self, key, [Emoji(i, self.__token) for i in json[key]])
            elif key == "joined_at":
                setattr(self, key, parse(json[key]) if json[key] else None)
            elif key == "voice_states":
                setattr(self, key, [VoiceState(i, self.__token)
                                    for i in json[key]])
            elif key == "members":
                setattr(self, key, [Member(i, self.__token)
                                    for i in json[key]])
            elif key == "channels":
                setattr(self, key, [Channel(i, self.__token)
                                    for i in json[key]])
            elif key == "threads":
                setattr(self, key, [Channel(i, self.__token)
                                    for i in json[key]])
            else:
                setattr(self, key, json[key])

    async def edit(self, params=None):
        """Edit a guild with API params."""
        if params is None:
            params = {}
        raise_error(params, "params", dict)

        atom, result = await Request().send_async_request(f"/guilds/{self.id}", "PATCH", self.__token, params)

        if atom == 0:
            return Guild(result, self.__token)
        else:
            raise EditGuildFailed(result)

    async def fetch_channels(self):
        """Fetch all channels in the guild and returns."""

        from .Channel import Channel

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/channels", "GET", self.__token)

        if atom == 0:
            return [Channel(i, self.__token) for i in result]
        else:
            raise FetchGuildChannelsFailed(result)

    async def create_channel(self, params=None):
        """Create new channel with API params."""
        if params is None:
            params = {}
        raise_error(params, "params", dict)

        from .Channel import Channel

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/channels", "POST", self.__token, params)

        if atom == 0:
            return Channel(result, self.__token)
        else:
            raise CreateGuildChannelFailed(result)

    async def edit_channel_position(self, *args):
        """Edit multiple channel position with API params."""

        for arg in args:
            raise_error(arg, "arg", dict)

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/channels", "PATCH", self.__token, [*args])

        if atom == 0:
            return True
        else:
            raise ChangeChannelPositionFailed(result)

    async def fetch_member(self, user_id: str):
        """Fetch a member from guild with id."""
        raise_error(user_id, "user_id", str)

        from .Member import Member

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/members/{user_id}", "GET", self.__token)

        if atom == 0:
            return Member(result, self.__token)
        else:
            raise FetchGuildMemberFailed(result)

    async def fetch_member_list(self, options: dict = None):
        """Fetch member list with API params."""
        if options is None:
            options = {}

        raise_error(options, "options", dict)

        from .Member import Member

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/members{d2q_converter(options)}", "GET",
                                                          self.__token)

        if atom == 0:
            return [Member(i, self.__token) for i in result]
        else:
            raise FetchGuildMembersFailed(result)

    async def search_members(self, options: dict = None):
        """Search members with API params (https://discord.com/developers/docs/resources/guild#search-guild-members-query-string-params)."""
        if options is None:
            options = {}

        raise_error(options, "options", dict)

        from .Member import Member

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/members/search{d2q_converter(options)}",
                                                          "GET", self.__token)

        if atom == 0:
            return [Member(i, self.__token) for i in result]
        else:
            raise SearchGuildMemberFailed(result)

    async def edit_member(self, member_id: str, options: dict = None):
        """Edit guild member with API params."""
        if options is None:
            options = {}

        raise_error(member_id, "member_id", str)
        raise_error(options, "options", dict)

        from .Member import Member

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/members/{member_id}", "PATCH",
                                                          self.__token, options)

        if atom == 0:
            return Member(result, self.__token)
        else:
            raise EditGuildMemberFailed(result)

    async def fetch_emoji_list(self):
        """Fetch all emojis in the guild."""

        from .Emoji import Emoji

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/emojis", "GET", self.__token)

        if atom == 0:
            return [Emoji(i, self.__token) for i in result]
        else:
            raise FetchGuildEmojiListFailed(result)

    async def fetch_emoji(self, emoji_id: str):
        """Fetch guild emoji with id."""

        from .Emoji import Emoji

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/emojis/{emoji_id}", "GET", self.__token)

        if atom == 0:
            return Emoji(result, self.__token)
        else:
            raise FetchGuildEmojiFailed(result)

    async def create_emoji(self, options=None):
        """Create guild emoji with API params."""

        if options is None:
            options = {}

        raise_error(options, "options", dict)

        from .Emoji import Emoji

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/emojis", "POST", self.__token, options)

        if atom == 0:
            return Emoji(result, self.__token)
        else:
            raise CreateGuildEmojiFailed(result)

    async def edit_emoji(self, emoji_id: str, options=None):
        """Edit guild emoji with API params."""
        if options is None:
            options = {}

        raise_error(emoji_id, "emoji_id", str)
        raise_error(options, "options", dict)

        from .Emoji import Emoji

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/emojis/{emoji_id}", "PATCH", self.__token,
                                                          options)

        if atom == 0:
            return Emoji(result, self.__token)
        else:
            raise EditGuildEmojiFailed(result)

    async def remove_emoji(self, emoji_id: str):
        """Remove guild emoji with id."""

        raise_error(emoji_id, "emoji_id", str)

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/emojis/{emoji_id}", "DELETE",
                                                          self.__token)

        if atom == 0:
            return True
        else:
            raise DeleteGuildEmojiFailed(result)

    async def add_role(self, member_id: str, role_id: str):
        """Add a role to guild member."""

        raise_error(member_id, "member_id", str)
        raise_error(role_id, "role_id", str)

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/members/{member_id}/roles/{role_id}",
                                                          "PUT", self.__token)

        if atom == 0:
            return self, member_id, role_id
        else:
            raise AddRoleToGuildMemberFailed(result)

    async def remove_role(self, member_id: str, role_id: str):
        """Remove a role from guild member."""

        raise_error(member_id, "member_id", str)
        raise_error(role_id, "role_id", str)

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/members/{member_id}/roles/{role_id}",
                                                          "DELETE", self.__token)

        if atom == 0:
            return self, member_id, role_id
        else:
            raise RemoveRoleFromGuildMemberFailed(result)

    async def kick(self, member_id: str):
        """Kick a member from guild."""

        raise_error(member_id, "member_id", str)

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/members/{member_id}", "DELETE",
                                                          self.__token)

        if atom == 0:
            return self, member_id
        else:
            raise KickMemberFromGuildFailed(result)

    async def ban(self, member_id: str, params: dict = None):
        """Ban a member from guild."""

        if params is None:
            params = {}

        raise_error(member_id, "member_id", str)
        raise_error(params, "params", dict)

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/bans/{member_id}", "PUT", self.__token,
                                                          params)

        if atom == 0:
            return self, member_id
        else:
            raise BanMemberFromGuildFailed(result)

    async def unban(self, member_id: str):
        """Unban a member from guild."""

        raise_error(member_id, "member_id", str)

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/bans/{member_id}", "DELETE", self.__token,
                                                          {})

        if atom == 0:
            return self, member_id
        else:
            raise UnbanGuildMemberFailed(result)

    async def fetch_bans(self):
        """Fetch all guild bans."""

        from .Ban import Ban

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/bans", "GET", self.__token)

        if atom == 0:
            return [Ban(i, self.__token) for i in result]
        else:
            raise FetchGuildBansFailed(result)

    async def fetch_member_ban(self, member_id: str):
        """Fetch an user guild ban."""
        raise_error(member_id, "member_id", str)

        from .Ban import Ban

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/bans/{member_id}", "GET", self.__token)

        if atom == 0:
            return Ban(result, self.__token)
        else:
            raise FetchGuildBanFailed(result)

    async def is_banned(self, member_id: str) -> bool:
        """Check is guild member banned from guild."""
        raise_error(member_id, "member_id", str)

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/bans/{member_id}", "GET", self.__token)

        if atom == 0:
            return True
        else:
            return False

    async def fetch_roles(self):
        """Fetch all roles from guild."""

        from .Role import Role

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/roles", "GET", self.__token)

        if atom == 0:
            return [Role(i) for i in result]
        else:
            return FetchGuildRolesFailed(result)

    async def create_role(self, params=None):
        """Create new guild role with API params."""

        from .Role import Role

        if params is None:
            params = {}

        raise_error(params, "params", dict)

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/roles", "POST", self.__token, params)

        if atom == 0:
            return Role(result)
        else:
            return CreateGuildRoleFailed(result)

    async def edit_role(self, role_id: str, params=None):
        """Edit a guild role with API params."""

        from .Role import Role

        if params is None:
            params = {}

        raise_error(role_id, "role_id", str)
        raise_error(params, "params", dict)

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/roles/{role_id}", "PATCH", self.__token,
                                                          params)

        if atom == 0:
            return Role(result)
        else:
            return EditGuildRoleFailed(result)

    async def edit_role_position(self, *args):
        """Edit a guild role with API params."""

        from .Role import Role

        for index, arg in enumerate(args):
            raise_error(arg, f"args #{index}", dict)

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/roles", "PATCH", self.__token, [*args])

        if atom == 0:
            return [Role(i) for i in result]
        else:
            return EditGuildRoleFailed(result)

    async def delete_role(self, role_id: str):
        """Delete a guild role."""

        raise_error(role_id, "role_id", str)

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/roles/{role_id}", "DELETE", self.__token,
                                                          {})

        if atom == 0:
            return self, role_id
        else:
            return DeleteGuildRoleFailed(result)

    async def change_nickname(self, params=None):
        """Change client-user guild nickname with API params."""
        if params is None:
            params = {}

        raise_error(params, "params", dict)

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/members/@me/nick", "PATCH", self.__token,
                                                          params)

        if atom == 0:
            return result
        else:
            return ChangeClientUserNicknameFailed(result)

    async def fetch_url(self):
        """Fetch guild vanity url."""

        from .Invite import Invite

        atom, result = await Request().send_async_request(f"/guilds/{self.id}/vanity-url", "GET", self.__token)

        if atom == 0:
            return Invite(result, self.__token)
        else:
            return FetchGuildVanityUrlFailed(result)



from dataclasses import dataclass


@dataclass
class GuildPreview:
    # Attrs
    def __init__(self, json, token) -> None:
        self.__token = token

        from .Emoji import Emoji

        for key in json:
            if key == "emojis":
                setattr(self, key, [Emoji(i, self.__token) for i in json[key]])
            else:
                setattr(self, key, json[key])



class EditGuildFailed(Exception):
    """Raises when editing the guild is failed."""

    pass


class FetchGuildChannelsFailed(Exception):
    """Raises when fetching the guild channels is failed."""

    pass


class CreateGuildChannelFailed(Exception):
    """Raises when creating new guild channel is failed."""

    pass


class ChangeChannelPositionFailed(Exception):
    """Raises when changing a channel position is failed."""

    pass


class FetchGuildMemberFailed(Exception):
    """Raises when fetching a guild member is failed."""

    pass


class FetchGuildMembersFailed(Exception):
    """Raises when fetching guild member list is failed."""

    pass


class SearchGuildMemberFailed(Exception):
    """Raises when searching the guild members is failed."""

    pass


class EditGuildMemberFailed(Exception):
    """Raises when editing the guild member is failed."""

    pass


class FetchGuildEmojiListFailed(Exception):
    """Raises when fetching the emoji list is failed."""

    pass


class FetchGuildEmojiFailed(Exception):
    """Raises when fetching the guild emoji is failed."""

    pass


class CreateGuildEmojiFailed(Exception):
    """Raises when creating a guild emoji is failed."""

    pass


class EditGuildEmojiFailed(Exception):
    """Raises when editing a guild emoji is failed."""

    pass


class DeleteGuildEmojiFailed(Exception):
    """Raises when delete a guild emoji is failed."""

    pass


class AddRoleToGuildMemberFailed(Exception):
    """Raises when adding a role to guild member is failed."""

    pass


class RemoveRoleFromGuildMemberFailed(Exception):
    """Raises when removing a role from guild member is failed."""

    pass


class KickMemberFromGuildFailed(Exception):
    """Raises when kicking a member from guild is failed."""

    pass


class BanMemberFromGuildFailed(Exception):
    """Raises when banning a member from guild is failed."""

    pass


class UnbanGuildMemberFailed(Exception):
    """Raises when unbanning a user from guild is failed."""

    pass


class FetchGuildBansFailed(Exception):
    """Raises when fetching the guild bans is failed."""

    pass


class FetchGuildBanFailed(Exception):
    """Raises when fetching the member guild ban is failed."""

    pass


class FetchGuildRolesFailed(Exception):
    """Raises when fetching the guild roles is failed."""

    pass


class CreateGuildRoleFailed(Exception):
    """Raises when creating new guild role is failed."""

    pass


class EditGuildRoleFailed(Exception):
    """Raises when editing the guild role is failed."""

    pass


class DeleteGuildRoleFailed(Exception):
    """Raises when deleting the guild role is failed."""

    pass


class ChangeClientUserNicknameFailed(Exception):
    """Raises when changing the client-user guild nickname is failed."""

    pass


class FetchGuildVanityUrlFailed(Exception):
    """Raises when fetching the vanity url is failed."""

    pass



from dataclasses import dataclass
from typing import Union


@dataclass
class Sticker:
    # Attrs
    def __init__(self, json) -> None:
        self.id: str = json['id']
        self.pack_id: str = json['pack_id']
        self.name: str = json['name']
        self.description: str = json['description']
        self.tags: Union[str, None] = json['tags'] if 'tags' in json else None
        self.asset: Union[str,
                          None] = json['asset'] if 'asset' in json else None
        self.format_type: int = json['format_type']



from dataclasses import dataclass


@dataclass
class Team:
    # Attrs
    def __init__(self, json, token) -> None:
        self.__token: str = token

        from .TeamMember import TeamMember

        for key in json:
            if key == "members":
                setattr(self, key, [TeamMember(i, self.__token)
                                    for i in json[key]])
            else:
                setattr(self, key, json[key])



from dataclasses import dataclass


@dataclass
class PresenceUpdate:
    # Attrs
    def __init__(self, json, token) -> None:
        self.__token: str = token

        from .User import User

        self.user: User = User(json['user'], self.__token)
        self.guild_id: str = json['guild_id']
        self.status: str = json['status']
        self.activities: list = json['activities']
        self.client_status: dict = json['client_status']



CREATE_INSTANT_INVITE = 0x0000000001  # Allows creation of instant invites
KICK_MEMBERS = 0x0000000002  # Allows kicking members
BAN_MEMBERS = 0x0000000004  # Allows banning members
ADMINISTRATOR = 0x0000000008  # Allows all permissions and bypasses channel permission overwrites
MANAGE_CHANNELS = 0x0000000010  # Allows management and editing of channels
MANAGE_GUILD = 0x0000000020  # Allows management and editing of the guild
ADD_REACTIONS = 0x0000000040  # Allows for the addition of reactions to messages
VIEW_AUDIT_LOG = 0x0000000080  # Allows for viewing of audit logs
PRIORITY_SPEAKER = 0x0000000100  # Allows for using priority speaker in a voice channel
STREAM = 0x0000000200  # Allows the user to go live
VIEW_CHANNEL = 0x0000000400  # Allows guild members to view a channel, which includes reading messages in text channels
SEND_MESSAGES = 0x0000000800  # Allows for sending messages in a channel
SEND_TTS_MESSAGES = 0x0000001000  # Allows for sending of /tts messages
MANAGE_MESSAGES = 0x0000002000  # Allows for deletion of other users messages
EMBED_LINKS = 0x0000004000  # Links sent by users with this permission will be auto-embedded
ATTACH_FILES = 0x0000008000  # Allows for uploading images and files
READ_MESSAGE_HISTORY = 0x0000010000  # Allows for reading of message history
MENTION_EVERYONE = 0x0000020000  # Allows for using the @everyone tag to notify all users in a channel,
# and the @here tag to notify all online users in a channel
USE_EXTERNAL_EMOJIS = 0x0000040000  # Allows the usage of custom emojis from other servers
VIEW_GUILD_INSIGHTS = 0x0000080000  # Allows for viewing guild insights
CONNECT = 0x0000100000  # Allows for joining of a voice channel
SPEAK = 0x0000200000  # Allows for speaking in a voice channel
MUTE_MEMBERS = 0x0000400000  # Allows for muting members in a voice channel
DEAFEN_MEMBERS = 0x0000800000  # Allows for deafening of members in a voice channel
MOVE_MEMBERS = 0x0001000000  # Allows for moving of members between voice channels
USE_VAD = 0x0002000000  # Allows for using voice-activity-detection in a voice channel
CHANGE_NICKNAME = 0x0004000000  # Allows for modification of own nickname
MANAGE_NICKNAMES = 0x0008000000  # Allows for modification of other users nicknames
MANAGE_ROLES = 0x0010000000  # Allows management and editing of roles
MANAGE_WEBHOOKS = 0x0020000000  # Allows management and editing of webhooks
MANAGE_EMOJIS = 0x0040000000  # Allows management and editing of emojis
USE_SLASH_COMMANDS = 0x0080000000  # Allows members to use slash commands in text channels
REQUEST_TO_SPEAK = 0x0100000000  # Allows for requesting to speak in stage channels.
# (This permission is under active development and may be changed or removed.)
MANAGE_THREADS = 0x0400000000  # Allows for deleting and archiving threads, and viewing all private threads
USE_PUBLIC_THREADS = 0x0800000000  # Allows for creating and participating in threads
USE_PRIVATE_THREADS = 0x1000000000  # Allows for creating and participating in private threads



import asyncio
import json
import sys
import traceback
from typing import Callable

import aiohttp

from ..errors.SessionErrors import TokenNotFoundError, InvalidTokenError, IntentNotFoundError
from ..models.Channel import Channel
from ..models.ClientUser import ClientUser
from ..models.Guild import Guild
from ..models.Member import Member
from ..models.Message import Message
from ..models.User import User
from ..utils.Request import Request
from ..utils.WrongType import raise_error

async_request = Request().send_async_request


class Session:
    """Session class for connection and etc.."""
    DISPATCH = 0
    HEARTBEAT = 1
    IDENTIFY = 2
    PRESENCE = 3
    VOICE_STATE = 4
    VOICE_PING = 5
    RESUME = 6
    RECONNECT = 7
    REQUEST_MEMBERS = 8
    INVALID_SESSION = 9
    HELLO = 10
    HEARTBEAT_ACK = 11
    GUILD_SYNC = 12

    def __init__(self, options: dict) -> None:
        if not isinstance(options, dict):
            raise TypeError("Options argument must be a dictionary.")

        if "TOKEN" not in options:
            raise TokenNotFoundError("Please pass a token in session option.")

        if "INTENTS" not in options:
            raise IntentNotFoundError(
                "Please pass a intent.")

        raise_error(options['TOKEN'], "TOKEN", str)
        raise_error(options['INTENTS'], "INTENTS", int)

        if 'MAX_MESSAGES' in options:
            raise_error(options['MAX_MESSAGES'], "MAX_MESSAGES", int)
            self.max_messages = options['MAX_MESSAGES']
        else:
            self.max_messages = 100

        self.token = options['TOKEN']
        self.intents = options['INTENTS']
        self.gateway = "wss://gateway.discord.gg/?v=9&encoding=json"
        self.ws = None
        self.client = None
        self.session_id = None
        self.event_loop = asyncio.get_event_loop()

        self.events = []
        self.__will_loaded_events = []

    def event(self, event_name: str, fn: Callable) -> True:
        """Create new event."""

        raise_error(event_name, "Event Name", str)
        raise_error(fn, "Function", type(lambda: True))

        self.__will_loaded_events.append({
            "EVENT": event_name,
            "FUNCTION": fn
        })

        return True

    async def __check_token(self):
        atom, result = await async_request("/users/@me", "GET", self.token)

        # Atom Check:
        if atom == 1:
            raise InvalidTokenError("Token is invalid. Please check your token!")
        else:
            self.client = ClientUser(result, self.token, self.max_messages)

    def __load_events(self):
        cache_events = (
            {
                "EVENT": "MESSAGE_CREATE",
                "FUNCTION": self.client.add_message_cache
            },
            {
                "EVENT": "MESSAGE_DELETE",
                "FUNCTION": self.client.remove_message_cache
            },
            {
                "EVENT": "MESSAGE_UPDATE",
                "FUNCTION": self.client.update_message_cache
            },
            {
                "EVENT": "MESSAGE_DELETE_BULK",
                "FUNCTION": self.client.bulk_delete_message_cache
            },
            {
                "EVENT": "GUILD_CREATE",
                "FUNCTION": self.client.add_guild_cache
            },
            {
                "EVENT": "GUILD_UPDATE",
                "FUNCTION": self.client.update_guild_cache
            },
            {
                "EVENT": "GUILD_DELETE",
                "FUNCTION": self.client.remove_guild_cache
            },
            {
                "EVENT": "GUILD_MEMBER_ADD",
                "FUNCTION": self.client.add_guild_member
            },
            {
                "EVENT": "GUILD_MEMBER_REMOVE",
                "FUNCTION": self.client.remove_guild_member
            },
            {
                "EVENT": "GUILD_MEMBER_UPDATE",
                "FUNCTION": self.client.update_guild_member
            },
            {
                "EVENT": "CHANNEL_CREATE",
                "FUNCTION": self.client.add_guild_channel
            },
            {
                "EVENT": "CHANNEL_UPDATE",
                "FUNCTION": self.client.update_guild_channel
            },
            {
                "EVENT": "CHANNEL_DELETE",
                "FUNCTION": self.client.remove_guild_channel
            },
            {
                "EVENT": "READY",
                "FUNCTION": self.__get_session_id
            }
        )

        # Load Cache System
        self.events.extend([
            *cache_events,
            *self.__will_loaded_events
        ])

    async def __connect_to_gateway(self):
        self.ws = await aiohttp.ClientSession().ws_connect(self.gateway)

    async def __get_session_id(self, packet):
        self.session_id = packet.get('session_id')

    async def __identify(self, packet):
        heartbeat = packet['d']['heartbeat_interval']

        if self.session_id is None:
            await self.ws.send_json({
                "op": self.IDENTIFY,
                "d": {
                    "token": self.token,
                    "intents": self.intents,
                    "properties": {
                        "$os": "linux",
                        "$browser": "5elenay/hyaline",
                        "$device": "5elenay/hyaline"
                    }
                }
            })
        else:
            await self.ws.send_json({
                "op": self.RESUME,
                "d": {
                    "token": self.token,
                    "session_id": self.session_id
                }
            })

        # Keep Connection Alive:
        async def _keep_alive():
            while True:
                await self.ws.send_json({
                    "op": self.HEARTBEAT,
                    "d": None
                })

                await asyncio.sleep(heartbeat / 1000)

        asyncio.run_coroutine_threadsafe(_keep_alive(), self.event_loop)

    def __filter_events(self, events: tuple = (), *args):
        return [
            event['FUNCTION'](*args) for event in self.events if event['EVENT'] in events
        ]

    async def __handle_event(self, packet):
        event_type = packet['t']
        event_data = packet['d']

        if event_type in [event['EVENT'] for event in self.events]:
            if event_type in ("MESSAGE_CREATE", "MESSAGE_UPDATE"):
                filtered = self.__filter_events((event_type,), Message(event_data, self.token))
            elif event_type in ("GUILD_CREATE", "GUILD_UPDATE"):
                filtered = self.__filter_events((event_type,), Guild(event_data, self.token))
            elif event_type in ("GUILD_MEMBER_ADD", "GUILD_MEMBER_UPDATE"):
                filtered = self.__filter_events((event_type,), event_data['guild_id'], Member(event_data, self.token))
            elif event_type == "GUILD_MEMBER_REMOVE":
                filtered = self.__filter_events((event_type,), event_data['guild_id'],
                                                User(event_data['user'], self.token))
            elif event_type in ("GUILD_BAN_ADD", "GUILD_BAN_REMOVE"):
                guild_id = event_data['guild_id']
                del event_data['guild_id']

                filtered = self.__filter_events((event_type,), guild_id, User(event_data, self.token))
            elif event_type in ("CHANNEL_CREATE", "CHANNEL_UPDATE", "CHANNEL_DELETE"):
                filtered = self.__filter_events((event_type,), Channel(event_data, self.token))
            else:
                filtered = self.__filter_events((event_type,), event_data)

            await asyncio.gather(*filtered)
        else:
            return None

    async def __receive(self):
        while True:
            packet = await self.ws.receive()

            # WebSocket Error
            if isinstance(packet.data, int) and len(str(packet.data)) == 4:
                print("WebSocket Exception Found: {0} ({1})".format(packet.data, packet.extra))
                continue
            elif isinstance(packet.data, type(None)):
                # WebSocket Closed
                if packet.type == 0x101:
                    return 0x0

            packet = json.loads(packet.data)
            # print(packet)

            if packet['op'] == self.HELLO:
                await self.__identify(packet)
            elif packet['op'] == self.RECONNECT:
                return 0x1
            elif packet['op'] == self.DISPATCH:
                try:
                    self.event_loop.create_task(self.__handle_event(packet))
                except Exception as error:
                    error = getattr(error, 'original', error)
                    print('Exception Found In Event {0}:'.format(packet['t']), file=sys.stderr)
                    traceback.print_exception(type(error), error, error.__traceback__, file=sys.stderr)

    async def __start_client(self):
        await self.__check_token()
        self.__load_events()
        await self.__connect_to_gateway()
        result = await self.__receive()

        # Reconnect
        if result == 0x1:
            await self.__start_client()

    async def change_presence(self, params: dict = None):
        if params is None:
            params = {}

        raise_error(params, "params", dict)

        await self.ws.send_json({
            "op": self.PRESENCE,
            "d": params
        })

    def start(self):
        """Start the session."""

        self.event_loop.run_until_complete(self.__start_client())



from setuptools import setup, find_packages

classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Operating System :: OS Independent',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Programming Language :: Python :: 3'
]

setup(
    name='hyaline',
    version='0.0.4',
    description='Discord API Wrapper for Python!',
    long_description_content_type="text/markdown",
    long_description=open('README.md').read(),
    url='https://github.com/5elenay/hyaline/',
    author='5elenay',
    author_email='',
    license='MIT',
    project_urls={
        "Bug Tracker": "https://github.com/5elenay/hyaline/issues",
    },
    classifiers=classifiers,
    keywords=["discord", "wrapper", "api", "gateway", "lightweight", "discord-api"],
    packages=find_packages(),
    install_requires=['python-dateutil', 'aiohttp']
)



def raise_error(variable, name, check) -> None:
    if isinstance(check, tuple):
        if not type(variable) in check:
            raise TypeError(
                f"Argument '{name}' type must be {' / '.join([i.__name__ if i != None else 'None' for i in check])}, not {type(variable).__name__}.")
    else:
        if not isinstance(variable, check):
            raise TypeError(
                f"Argument '{name}' type must be {check.__name__ if check != None else 'None'}, not {type(variable).__name__}.")



from dataclasses import dataclass
from typing import Union


@dataclass
class MessageActivity:
    # Attrs
    def __init__(self, json) -> None:
        self.type: int = json['type']
        self.party_id: Union[str,
                             None] = json['party_id'] if 'party_id' in json else None



from .Activity import *
from .Application import *
from .Attachment import *
from .Channel import *
from .ClientUser import *
from .Embed import *
from .Emoji import *
from .Guild import *
from .GuildPreview import *
from .Invite import *
from .Member import *
from .Message import *
from .MessageActivity import *
from .MessageReference import *
from .Overwrite import *
from .PresenceUpdate import *
from .Reaction import *
from .Role import *
from .RoleTags import *
from .Sticker import *
from .Team import *
from .TeamMember import *
from .ThreadMember import *
from .ThreadMetadata import *
from .User import *
from .VoiceState import *



from dataclasses import dataclass


@dataclass
class Emoji:
    # Attrs
    def __init__(self, json, token) -> None:
        self.__token: str = token

        from .User import User

        for key in json:
            if key == "user":
                setattr(self, key, User(json[key], self.__token))
            else:
                setattr(self, key, json[key])



from dataclasses import dataclass


@dataclass
class Overwrite:
    # Attrs
    def __init__(self, json) -> None:
        self.id: str = json['id']
        self.type: int = json['type']
        self.allow: str = json['allow']
        self.deny: str = json['deny']



from dataclasses import dataclass
from typing import Union


@dataclass
class MessageReference:
    # Attrs
    def __init__(self, json) -> None:
        self.message_id: Union[str,
                               None] = json['message_id'] if 'message_id' in json else None
        self.channel_id: Union[str,
                               None] = json['channel_id'] if 'channel_id' in json else None
        self.guild_id: Union[str,
                             None] = json['guild_id'] if 'guild_id' in json else None
        self.fail_if_not_exists: Union[bool,
                                       None] = json['fail_if_not_exists'] if 'fail_if_not_exists' in json else None



# ChannelTypes
from dataclasses import dataclass


@dataclass
class Channel:
    GUILD_TEXT = 0  # a text channel within a server
    DM = 1  # a direct message between users
    GUILD_VOICE = 2  # a voice channel within a server
    GROUP_DM = 3  # a direct message between multiple users
    GUILD_CATEGORY = 4  # an organizational category that contains up to 50 channels
    GUILD_NEWS = 5  # a channel that users can follow and cross post into their own server
    GUILD_STORE = 6  # a channel in which game developers can sell their game on Discord
    GUILD_NEWS_THREAD = 10  # a temporary sub-channel within a GUILD_NEWS channel
    GUILD_PUBLIC_THREAD = 11  # a temporary sub-channel within a GUILD_TEXT channel
    GUILD_PRIVATE_THREAD = 12  # a temporary sub-channel within a GUILD_TEXT channel that is only viewable by
    # those invited and those with the MANAGE_THREADS permission
    GUILD_STAGE_VOICE = 13  # a voice channel for hosting events with an audience


@dataclass
class Activity:
    GAMING = 0
    STREAMING = 1
    LISTENING = 2
    WATCHING = 3
    CUSTOM = 4
    COMPETING = 5
